import json
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Self

from jax import numpy as jnp
from jaxtyping import Array, DTypeLike

from lalamo.common import ParameterPath
from lalamo.model_import.loaders.common import load_parameters
from lalamo.model_import.loaders.fishaudio_loaders import (
    load_fishaudio_audio_decoder,
    load_fishaudio_text_decoder,
)
from lalamo.model_import.model_configs import ForeignTTSConfig
from lalamo.modules import (
    GELU,
    AttentionConfig,
    DenseMLPConfig,
    FullPrecisionLinearConfig,
    LalamoModule,
    NormalizationConfig,
    SiLU,
    TiedEmbeddingConfig,
    TransformerConfig,
    TransformerLayerConfig,
    TTSConfig,
    TTSModel,
    UnscaledRoPEConfig,
    UpcastMode,
    VocoderConfig,
)
from lalamo.modules.audio.common_modules import (
    CausalConv1dConfig,
)
from lalamo.modules.audio.fishaudio import (
    DescriptAudioCodec,
    DescriptAudioCodecConfig,
    FishAudioTextDecoder,
    FishAudioTextDecoderConfig,
)
from lalamo.modules.audio.fishaudio.fishaudio_common import get_default_fishaudio_dac_config
from lalamo.modules.audio.fishaudio.fishaudio_modules import (
    CausalTransposeConv1dConfig,
    ConvNeXtBlockConfig,
    DACDecoderBlockConfig,
    DACDecoderConfig,
    DownsampleResidualVectorQuantizeConfig,
    ResidualUnitConfig,
    ResidualVectorQuantizeConfig,
    Snake1dConfig,
    UpsamplerConfig,
    UpsamplingBlockConfig,
    VectorQuantizeConfig,
)

__all__ = ["FishAudioConfig"]


def lalamo_transformer_cfg_from_fish_audio_codec_cfg(
    config: Mapping[Any, Any],
    precision: DTypeLike,
    accumulation_precision: DTypeLike,
    window_size: int,
    input_dim: int,
) -> TransformerConfig:
    # NOTE: this condifion is from post_init() for the post-module config object
    n_local_heads = config["n_head"] if config["n_local_heads"] == -1 else config["n_local_heads"]

    global_rope_config = UnscaledRoPEConfig(
        precision=precision,
        base=config["rope_base"],
        max_sequence_length=config["block_size"],
    )
    local_rope_config = None

    norm_config_pre = NormalizationConfig(
        scale_precision=precision,
        accumulation_precision=accumulation_precision,
        epsilon=config["norm_eps"],
        scale_offset=None,
        upcast_mode=UpcastMode.ONLY_NORMALIZATION,
        subtract_mean=False,
    )

    qkv_projection_config = FullPrecisionLinearConfig(precision=precision)
    out_projection_config = FullPrecisionLinearConfig(precision=precision)
    mixer_config = AttentionConfig(
        qkv_projection_config=qkv_projection_config,
        out_projection_config=out_projection_config,
        query_norm_config=None,
        key_norm_config=None,
        num_heads=config["n_head"],
        num_groups=n_local_heads,
        head_dim=config["head_dim"],
        is_causal=True,
        scale=None,
        sliding_window_size=window_size,
        logit_soft_cap=None,
        has_sinks=False,
        has_qkv_biases=False,
        has_out_biases=False,
    )

    mlp_linear_config = FullPrecisionLinearConfig(precision=precision)
    mlp_use_up_biases = False
    mlp_use_down_biases = False
    mlp_config = DenseMLPConfig(
        linear_config=mlp_linear_config,
        activation=SiLU(),
        has_up_biases=mlp_use_up_biases,
        has_down_biases=mlp_use_down_biases,
        gate_clipping=None,
        up_clipping=None,
    )

    pre_mixer_norm_config = norm_config_pre
    post_mixer_norm_config = None
    pre_mlp_norm_config = norm_config_pre
    post_mlp_norm_config = None

    layer_config = TransformerLayerConfig(
        pre_mixer_norm_config=pre_mixer_norm_config,
        mixer_config=mixer_config,
        post_mixer_norm_config=post_mixer_norm_config,
        pre_mlp_norm_config=pre_mlp_norm_config,
        mlp_config=mlp_config,
        post_mlp_norm_config=post_mlp_norm_config,
    )
    hidden_dim = config["intermediate_size"]
    context_length = config["block_size"]

    transformer_cfg = TransformerConfig(
        global_rope_config=global_rope_config,
        local_rope_config=local_rope_config,
        layer_configs=tuple([layer_config] * config["n_layer"]),
        output_norm_config=norm_config_pre,
        model_dim=input_dim,
        hidden_dim=hidden_dim,
        context_length=context_length,
    )

    return transformer_cfg


def instantiate_dac_config_from_fishaudio_config(
    fish_dac_config: Mapping[Any, Any],
    precision: DTypeLike,
    accumulation_precision: DTypeLike,
) -> DescriptAudioCodecConfig:
    samplerate = fish_dac_config["sample_rate"]
    fish_quantizer_config = fish_dac_config["quantizer"]

    input_dim = fish_quantizer_config["input_dim"]
    downsample_factor = fish_quantizer_config["downsample_factor"]
    post_module_config_dict = fish_quantizer_config["post_module"]
    encoder_dim = fish_dac_config["encoder_dim"]
    encoder_rates = fish_dac_config["encoder_rates"]
    decoder_dim = fish_dac_config["decoder_dim"]
    decoder_rates = fish_dac_config["decoder_rates"]
    fish_quantizer_config = fish_dac_config["quantizer"]
    input_dim = fish_quantizer_config["input_dim"]
    n_codebooks = fish_quantizer_config["n_codebooks"]
    codebook_dim = fish_quantizer_config["codebook_dim"]
    downsample_factor = fish_quantizer_config["downsample_factor"]
    codebook_size = fish_quantizer_config["codebook_size"]
    semantic_codebook_size = fish_quantizer_config["semantic_codebook_size"]

    convnext_config = ConvNeXtBlockConfig(
        precision=precision,
        activation=GELU(approximate=False),
        dwconv_config=CausalConv1dConfig(precision=precision, has_biases=True),
        norm_config=NormalizationConfig(
            scale_precision=precision,
            accumulation_precision=accumulation_precision,
            epsilon=1e-6,
            scale_offset=None,
            upcast_mode=UpcastMode.FULL_LAYER,
            subtract_mean=True,
            use_bias=True,
        ),
        pwconv_config=FullPrecisionLinearConfig(precision=precision),
    )
    upsampling_block_config = UpsamplingBlockConfig(
        precision=precision,
        trans_conv_config=CausalTransposeConv1dConfig(precision=precision, has_biases=True),
        convnext_config=convnext_config,
    )
    num_blocks = len(downsample_factor)
    block_configs = tuple(upsampling_block_config for _ in range(num_blocks))
    upsampler_config = UpsamplerConfig(block_configs=block_configs)

    post_module_transformer_foreign = post_module_config_dict["config"]
    post_module_config = lalamo_transformer_cfg_from_fish_audio_codec_cfg(
        post_module_transformer_foreign,
        precision,
        accumulation_precision,
        window_size=post_module_config_dict["window_size"],
        input_dim=post_module_config_dict["input_dim"],
    )

    vq_config = VectorQuantizeConfig(
        precision=precision,
        codebook_config=TiedEmbeddingConfig(
            input_scale=None,
            logit_soft_cap=None,
            precision=precision,
        ),
        out_proj_config=FullPrecisionLinearConfig(precision=precision),
    )
    lalamo_rvq_config = ResidualVectorQuantizeConfig(
        precision=precision,
        vq_config=vq_config,
    )

    quantizer_full_config = DownsampleResidualVectorQuantizeConfig(
        precision=precision,
        semantic_quantizer_config=lalamo_rvq_config,
        quantizer_config=lalamo_rvq_config,
        post_module_config=post_module_config,
        upsampler_config=upsampler_config,
    )
    res_unit_config = ResidualUnitConfig(
        precision=precision,
        snake_config=Snake1dConfig(precision=precision),
        conv_config=CausalConv1dConfig(precision=precision, has_biases=True),
        causal=True,
    )
    decoder_block_config = DACDecoderBlockConfig(
        precision=precision,
        snake_config=Snake1dConfig(precision=precision),
        trans_conv_config=CausalTransposeConv1dConfig(precision=precision, has_biases=True),
        res_unit_config=res_unit_config,
        causal=True,
    )
    decoder_config = DACDecoderConfig(
        precision=precision,
        conv_config=CausalConv1dConfig(precision=precision, has_biases=True),
        snake_config=Snake1dConfig(precision=precision),
        decoder_block_config=decoder_block_config,
        causal=True,
    )

    return DescriptAudioCodecConfig(
        precision=precision,
        quantizer_config=quantizer_full_config,
        decoder_config=decoder_config,
        samplerate=samplerate,
        encoder_dim=encoder_dim,
        encoder_rates=encoder_rates,
        decoder_dim=decoder_dim,
        decoder_rates=decoder_rates,
        input_dim=input_dim,
        n_codebooks=n_codebooks,
        codebook_dim=codebook_dim,
        downsample_factor=downsample_factor,
        codebook_size=codebook_size,
        semantic_codebook_size=semantic_codebook_size,
    )


@dataclass(frozen=True)
class FishAudioConfig(ForeignTTSConfig):
    attention_o_bias: bool
    attention_qk_norm: bool
    attention_qkv_bias: bool
    codebook_size: int
    dim: int
    dropout: float
    fast_attention_o_bias: bool
    fast_attention_qk_norm: bool
    fast_attention_qkv_bias: bool
    fast_dim: int
    fast_head_dim: int
    fast_intermediate_size: int
    fast_n_head: int
    fast_n_local_heads: int
    head_dim: int
    initializer_range: int
    intermediate_size: int
    max_seq_len: int
    model_type: str
    n_fast_layer: int
    n_head: int
    n_layer: int
    n_local_heads: int
    norm_eps: float
    num_codebooks: int
    rope_base: int
    scale_codebook_embeddings: bool
    tie_word_embeddings: bool
    use_gradient_checkpointing: bool
    vocab_size: int

    # NOTE: these fields are used during inference but must be retrieved from
    # tokenizer config files
    semantic_token_begin_id: int = -1
    semantic_token_end_id: int = -1
    im_end_token_id: int = -1

    def extract_textual_transformer_configs(
        self,
        precision: DTypeLike,
        accumulation_precision: DTypeLike,
        fast_module: bool = False,
    ) -> tuple[TransformerConfig, FullPrecisionLinearConfig]:
        n_layer = self.n_fast_layer if fast_module else self.n_layer
        n_head = self.fast_n_head if fast_module else self.n_head
        dim = self.fast_dim if fast_module else self.dim
        intermediate_size = self.fast_intermediate_size if fast_module else self.intermediate_size
        n_local_heads = self.fast_n_local_heads if fast_module else self.n_local_heads
        head_dim = self.fast_head_dim if fast_module else self.head_dim
        attention_qk_norm = self.fast_attention_qk_norm if fast_module else self.attention_qk_norm

        global_rope_config = UnscaledRoPEConfig(
            precision=precision,
            base=self.rope_base,
            max_sequence_length=self.max_seq_len,
        )
        local_rope_config = None

        norm_config = NormalizationConfig(
            scale_precision=precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.norm_eps,
            scale_offset=None,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
            subtract_mean=False,
        )

        qkv_projection_config = FullPrecisionLinearConfig(precision=precision)
        out_projection_config = FullPrecisionLinearConfig(precision=precision)
        mixer_config = AttentionConfig(
            qkv_projection_config=qkv_projection_config,
            out_projection_config=out_projection_config,
            query_norm_config=norm_config if attention_qk_norm else None,
            key_norm_config=norm_config if attention_qk_norm else None,
            num_heads=n_head,
            num_groups=n_local_heads,
            head_dim=head_dim,
            is_causal=True,
            scale=None,
            sliding_window_size=None,
            logit_soft_cap=None,
            has_sinks=False,
            has_qkv_biases=False,
            has_out_biases=False,
        )

        mlp_linear_config = FullPrecisionLinearConfig(precision=precision)
        mlp_use_up_biases = False
        mlp_use_down_biases = False
        mlp_config = DenseMLPConfig(
            linear_config=mlp_linear_config,
            activation=SiLU(),
            has_up_biases=mlp_use_up_biases,
            has_down_biases=mlp_use_down_biases,
            gate_clipping=None,
            up_clipping=None,
        )

        pre_mixer_norm_config = norm_config
        post_mixer_norm_config = None
        pre_mlp_norm_config = norm_config
        post_mlp_norm_config = None

        layer_config = TransformerLayerConfig(
            pre_mixer_norm_config=pre_mixer_norm_config,
            mixer_config=mixer_config,
            post_mixer_norm_config=post_mixer_norm_config,
            pre_mlp_norm_config=pre_mlp_norm_config,
            mlp_config=mlp_config,
            post_mlp_norm_config=post_mlp_norm_config,
        )
        model_dim = dim
        hidden_dim = intermediate_size
        context_length = self.max_seq_len

        transformer_cfg = TransformerConfig(
            global_rope_config=global_rope_config,
            local_rope_config=local_rope_config,
            layer_configs=tuple([layer_config] * n_layer),
            output_norm_config=norm_config,
            model_dim=model_dim,
            hidden_dim=hidden_dim,
            context_length=context_length,
        )
        linear_out_cfg = FullPrecisionLinearConfig(precision=precision)

        return (transformer_cfg, linear_out_cfg)

    def to_tts_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
    ) -> TTSConfig:
        audio_decoder_config = instantiate_dac_config_from_fishaudio_config(
            fish_dac_config=get_default_fishaudio_dac_config(),
            precision=activation_precision,
            accumulation_precision=accumulation_precision,
        )

        slow_transformer_cfg, slow_readout_cfg = self.extract_textual_transformer_configs(
            precision=activation_precision,
            accumulation_precision=accumulation_precision,
            fast_module=False,
        )
        fast_transformer_cfg, fast_readout_cfg = self.extract_textual_transformer_configs(
            precision=activation_precision,
            accumulation_precision=accumulation_precision,
            fast_module=True,
        )
        slow_embedding_cfg = TiedEmbeddingConfig(input_scale=None, logit_soft_cap=None, precision=activation_precision)
        fast_embedding_cfg = TiedEmbeddingConfig(input_scale=None, logit_soft_cap=None, precision=activation_precision)

        codebook_embeddings_cfg = TiedEmbeddingConfig(
            input_scale=None,
            logit_soft_cap=None,
            precision=activation_precision,
        )
        if self.dim == self.fast_dim:
            fast_model_projection_config = None
        else:
            fast_model_projection_config = FullPrecisionLinearConfig(activation_precision)
        text_decoder_config = FishAudioTextDecoderConfig(
            slow_embeddings_config=slow_embedding_cfg,
            slow_model_config=slow_transformer_cfg,
            slow_readout_config=slow_readout_cfg,
            fast_embeddings_config=fast_embedding_cfg,
            fast_model_config=fast_transformer_cfg,
            fast_readout_config=fast_readout_cfg,
            codebook_embeddings_config=codebook_embeddings_cfg,
            fast_model_projection_config=fast_model_projection_config,
            semantic_token_begin_id=self.semantic_token_begin_id,
            semantic_token_end_id=self.semantic_token_end_id,
            im_end_token_id=self.im_end_token_id,
            codebook_size=self.codebook_size,
            vocab_size=self.vocab_size,
            slow_model_dim=self.dim,
            fast_model_dim=self.fast_dim,
            num_codebooks=self.num_codebooks,
            max_seq_len=min(context_length, self.max_seq_len) if context_length else self.max_seq_len,
            scale_codebook_embeddings=self.scale_codebook_embeddings,
            precision=activation_precision,
        )
        return TTSConfig(
            text_decoder_config=text_decoder_config,
            audio_decoder_config=audio_decoder_config,
            vocoder_config=VocoderConfig(),
            activation_precision=activation_precision,
        )

    def _load_weights(
        self,
        model: LalamoModule,
        weights_dict: Mapping[str, Array],
    ) -> LalamoModule:
        assert isinstance(model, TTSModel)

        assert isinstance(model.text_decoder, FishAudioTextDecoder)
        loaded_text_decoder = load_fishaudio_text_decoder(model.text_decoder, weights_dict, ParameterPath())

        assert isinstance(model.audio_decoder, DescriptAudioCodec)
        loaded_audio_decoder = load_fishaudio_audio_decoder(model.audio_decoder, weights_dict, ParameterPath())

        return load_parameters(
            lambda m: (
                m.text_decoder,
                m.audio_decoder,
            ),
            model,
            (loaded_text_decoder, loaded_audio_decoder),
        )

    @classmethod
    def from_json(cls, json_path: Path | str) -> Self:
        json_path = Path(json_path)
        with open(json_path) as f:
            config = json.load(f)
        return cls(**config)

    @property
    def default_precision(self) -> DTypeLike:
        # NOTE: in reality FishAudio text-decoder is bf16 while audio-decoder if fp32.
        # Currently lalamo weight manipulation pipeline does not support such
        # mixed-model-mixed-weight configuration so we upcast everything to fp32
        # as temporary solution
        return jnp.dtype(getattr(self, "torch_dtype", "float32"))



from collections.abc import Callable, Iterable
from itertools import chain
from typing import NamedTuple

from lalamo.data.lalamo_completions import LalamoCompletion
from lalamo.data.utils import get_prefixes_ending_in_user_message
from lalamo.message_processor import Message
from lalamo.models import LanguageModel
from lalamo.models.common import InferenceConfig


class CollectTracesEvent(NamedTuple):
    sequences_processed: int
    tokens_generated: int


def inference_collect_traces(
    model: LanguageModel,
    conversations: Iterable[Iterable[Message]],
    num_top_logits_to_collect: int = 8,
    batch_size: int = 1,
    max_input_length: int = 1024,
    max_output_length: int = 1024,
    tokens_to_generate: int | None = None,
    progress_callback: Callable[[CollectTracesEvent], None] | None = None,
) -> Iterable[LalamoCompletion]:
    prefixes = chain.from_iterable(map(get_prefixes_ending_in_user_message, conversations))
    tokenized_prefixes = map(model.message_processor.tokenize_request, prefixes)
    filtered_prefixes = filter(lambda conv: len(conv) <= max_input_length, tokenized_prefixes)
    filtered_prefixes = list(filtered_prefixes)  # eagerly materialize the prompts into RAM

    config = InferenceConfig(
        max_output_length=max_output_length,
        num_top_logits_to_return=num_top_logits_to_collect,
        padded_length=max_input_length,
        batch_size=batch_size,
    )

    tokens_generated = 0

    for idx, generated in enumerate(
        model.generate_tokens_many(
            filtered_prefixes,
            inference_config=config,
        ),
    ):
        token_ids = generated.token_ids.tolist()
        seqlen = next(
            (i + 1 for i, t in enumerate(token_ids) if t in model.stop_token_ids),
            len(token_ids),
        )

        if tokens_to_generate is not None:
            seqlen = min(seqlen, tokens_to_generate - tokens_generated)

        tokens_generated += seqlen

        token_ids = token_ids[:seqlen]

        assert generated.top_k_token_ids is not None and generated.top_k_token_logits is not None
        token_logits_ids = generated.top_k_token_ids[:seqlen].tolist()
        token_logits_values = generated.top_k_token_logits[:seqlen].tolist()
        token_logits = [
            dict(zip(keys, values, strict=True))
            for keys, values in zip(token_logits_ids, token_logits_values, strict=True)
        ]

        # We need the original prompt tokens - get from indexed_inputs
        yield LalamoCompletion(filtered_prefixes[idx], token_ids, token_logits)

        if progress_callback is not None:
            progress_callback(CollectTracesEvent(idx + 1, tokens_generated))

        if tokens_to_generate is not None and tokens_generated >= tokens_to_generate:
            break



from collections.abc import Mapping

import jax.numpy as jnp
from jaxtyping import Array

from lalamo.common import ParameterPath


def fuse_weight_norm_conv1d_as_linear(
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> tuple[Array, Array | None]:
    """Fuse weight normalization for a Conv1d layer (mimic PyTorch's remove_weight_norm).

    Here we expect conv-1d with kernel_size==1, so basically it is used as linear projection

    Args:
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Path to the weight-normalized layer

    Returns:
        Tuple of (fused_weight, bias) as JAX arrays.
    """
    weight_g = weights_dict[path / "weight_g"]
    weight_v = weights_dict[path / "weight_v"]
    bias = weights_dict[path / "bias"]

    weight = weight_g * (weight_v / jnp.linalg.norm(weight_v, axis=1, keepdims=True))
    return (weight, bias)


def fuse_parametrized_weight_norm_conv1d(
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    is_transposed: bool = False,
) -> tuple[Array, Array | None]:
    """Fuse weight normalization for a Conv1d layer (mimic PyTorch's remove_parametrizations).

    This handles the newer parametrization format where weights are stored as:
        - path/parametrizations/weight/original0 (weight_g)
        - path/parametrizations/weight/original1 (weight_v)
        - path/bias

    Args:
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Path to the weight-normalized layer.
        is_transposed: If True, creates ConvTranspose1d instead of Conv1d.

    Returns:
        Tuple of (fused_weight, bias) as JAX arrays.
    """
    weight_g = weights_dict[path / "parametrizations" / "weight" / "original0"]
    weight_v = weights_dict[path / "parametrizations" / "weight" / "original1"]
    bias = weights_dict[path / "bias"]

    if is_transposed:
        # ConvTranspose1d weight shape: (in_channels, out_channels, kernel_size)
        reshape_dim, _, _ = weight_v.shape
    else:
        # Conv1d weight shape: (out_channels, in_channels, kernel_size)
        reshape_dim, _, _ = weight_v.shape

    norms = jnp.linalg.norm(weight_v.reshape(reshape_dim, -1), axis=1, keepdims=True)
    norms = norms.reshape(reshape_dim, 1, 1)

    weight = weight_g * (weight_v / norms)
    return (weight, bias)



import jax.numpy as jnp
import mlx.core as mx  # type: ignore
from jaxtyping import Array

__all__ = ["jax_to_mlx", "mlx_to_jax"]


def mlx_to_jax(a: mx.array) -> Array:
    if a.dtype == mx.bfloat16:
        return jnp.asarray(a.view(mx.uint16)).view(jnp.bfloat16)

    return jnp.asarray(a)


def jax_to_mlx(a: Array) -> mx.array:
    if a.dtype == jnp.bfloat16:
        return mx.array(a.view(jnp.uint16)).view(mx.bfloat16)

    return mx.array(a)



from lalamo.model_import.model_configs import HFMistralConfig

from .common import (
    ConfigMap,
    FileSpec,
    JSONFieldSpec,
    ModelSpec,
    UseCase,
    WeightsType,
)

__all__ = ["MISTRAL_MODELS"]

CODESTRAL_TOKENIZER_REPO = "mistralai/Codestral-22B-v0.1"

CODESTRAL = [
    ModelSpec(
        vendor="Mistral",
        family="Codestral",
        name="Codestral-22B-v0.1",
        size="22B",
        quantization=None,
        repo="mistral-community/Codestral-22B-v0.1",
        config_type=HFMistralConfig,
        weights_type=WeightsType.SAFETENSORS,
        use_cases=(UseCase.CODE,),
        configs=ConfigMap(
            tokenizer_config=FileSpec(repo=CODESTRAL_TOKENIZER_REPO, filename="tokenizer_config.json"),
        ),
    ),
]


DEVSTRAL_TOKENIZER_REPO = "mistralai/Mistral-Small-3.1-24B-Base-2503"
DEVSTRAL_CHAT_REPO = "mistralai/Mistral-Small-3.1-24B-Instruct-2503"

DEVSTRAL = [
    ModelSpec(
        vendor="Mistral",
        family="Devstral",
        name="Devstral-Small-2505",
        size="24B",
        quantization=None,
        repo="mistralai/Devstral-Small-2505",
        config_type=HFMistralConfig,
        configs=ConfigMap(
            tokenizer=FileSpec(repo=DEVSTRAL_TOKENIZER_REPO, filename="tokenizer.json"),
            tokenizer_config=FileSpec(repo=DEVSTRAL_TOKENIZER_REPO, filename="tokenizer_config.json"),
            generation_config=FileSpec(repo=DEVSTRAL_TOKENIZER_REPO, filename="generation_config.json"),
            chat_template=JSONFieldSpec(
                FileSpec(repo=DEVSTRAL_CHAT_REPO, filename="chat_template.json"),
                "chat_template",
            ),
            system_prompt=FileSpec("SYSTEM_PROMPT.txt"),
        ),
        use_cases=(UseCase.CODE,),
    ),
]


MISTRAL_MODELS = CODESTRAL + DEVSTRAL



from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
import jax
import jax.numpy as jnp
from einops import einsum, rearrange
from jax import vmap
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree, dummy_array, require_array, require_mapping, require_tree
from lalamo.modules.activations import Activation
from lalamo.modules.common import PositionalEmbeddingSelector
from lalamo.modules.linear import LinearBase, LinearConfig
from lalamo.modules.rope import PositionalEmbeddings
from lalamo.modules.token_mixers.state.ssm_state import SSMStateLayer

from .common import TokenMixerBase, TokenMixerConfigBase, TokenMixerResult
from .convolutions import SeparableCausalConv, SeparableCausalConvConfig

__all__ = [
    "Mamba2",
    "Mamba2Config",
    "Mamba2Result",
    "SeparableCausalConv",
    "SeparableCausalConvConfig",
    "exp_segsum",
    "fused_ssd_intra_chunk",
]

Mamba2Result = TokenMixerResult[SSMStateLayer]


def exp_segsum(x: Float[Array, "... T"]) -> Float[Array, "... T T"]:
    """Compute exp(segsum(x)) as lower-triangular matrix using cumsum difference."""
    seq_len = x.shape[-1]
    cs = jnp.cumsum(x, axis=-1)
    diff = cs[..., :, None] - cs[..., None, :]
    mask = jnp.tril(jnp.ones((seq_len, seq_len), dtype=bool))
    return jnp.where(mask, jnp.exp(diff), 0.0)


def fused_ssd_intra_chunk(
    a_cumsum: Float[Array, "groups heads_per_group chunks chunk_size"],
    cb: Float[Array, "chunks chunk_size chunk_size groups"],
    x: Float[Array, "chunks chunk_size groups heads_per_group head_dim"],
) -> Float[Array, "chunks chunk_size groups heads_per_group head_dim"]:
    """Compute intra-chunk diagonal block outputs for SSD.

    Avoids materializing the full global L matrix by computing decay locally per (chunk, group, head).
    """
    groups, heads_per_group, chunks, chunk_size = a_cumsum.shape

    def compute_one(
        a_cs: Float[Array, " chunk_size"],
        cb_slice: Float[Array, "chunk_size chunk_size"],
        x_slice: Float[Array, "chunk_size head_dim"],
    ) -> Float[Array, "chunk_size head_dim"]:
        diff = a_cs[:, None] - a_cs[None, :]
        mask = jnp.tril(jnp.ones((chunk_size, chunk_size), dtype=jnp.bool_))
        decay_local = jnp.where(mask, jnp.exp(diff), 0.0)
        weighted = decay_local * cb_slice
        return weighted @ x_slice

    def compute_chunk_group_head(chunk_idx: int, group_idx: int, head_idx: int) -> Float[Array, "chunk_size head_dim"]:
        return compute_one(
            a_cumsum[group_idx, head_idx, chunk_idx, :],
            cb[chunk_idx, :, :, group_idx],
            x[chunk_idx, :, group_idx, head_idx, :],
        )

    def over_heads(chunk_idx: int, group_idx: int) -> Float[Array, "heads_per_group chunk_size head_dim"]:
        return jax.vmap(lambda head_idx: compute_chunk_group_head(chunk_idx, group_idx, head_idx))(
            jnp.arange(heads_per_group),
        )

    def over_groups(chunk_idx: int) -> Float[Array, "groups heads_per_group chunk_size head_dim"]:
        return jax.vmap(lambda group_idx: over_heads(chunk_idx, group_idx))(jnp.arange(groups))

    result = jax.vmap(over_groups)(jnp.arange(chunks))

    return rearrange(
        result,
        "chunks groups heads_per_group chunk_size head_dim -> chunks chunk_size groups heads_per_group head_dim",
    )


@dataclass(frozen=True)
class Mamba2Config(TokenMixerConfigBase):
    in_projection_config: LinearConfig
    out_projection_config: LinearConfig
    conv_config: SeparableCausalConvConfig
    activation: Activation

    kernel_size: int
    num_heads: int
    num_groups: int
    head_dim: int
    state_dim: int
    expansion_factor: int

    has_in_biases: bool
    has_out_biases: bool

    chunk_size: int = 256

    @property
    def inner_dim(self) -> int:
        return self.num_heads * self.head_dim

    @property
    def conv_dim(self) -> int:
        return self.inner_dim + 2 * self.num_groups * self.state_dim

    @property
    def rope_dim(self) -> None:
        return None

    def random_init(
        self,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "Mamba2":
        in_key, out_key, conv_key, skip_key = jax.random.split(key, 4)

        in_projection = self.in_projection_config.random_init(
            input_dim=model_dim,
            output_dims=(
                self.inner_dim + 2 * self.num_groups * self.state_dim,
                self.inner_dim,
                self.num_heads,
            ),
            has_biases=self.has_in_biases,
            key=in_key,
        )

        out_projection = self.out_projection_config.random_init(
            self.inner_dim,
            (model_dim,),
            has_biases=self.has_out_biases,
            key=out_key,
        )

        conv = self.conv_config.random_init(self.conv_dim, self.kernel_size, key=conv_key)

        skip_connection_weight = jax.random.normal(
            skip_key,
            (self.num_heads,),
            dtype=in_projection.activation_precision,
        )

        gate_bias = jnp.zeros((self.inner_dim,), dtype=in_projection.activation_precision)

        return Mamba2(
            self,
            in_projection=in_projection,
            conv=conv,
            out_projection=out_projection,
            skip_connection_weight=skip_connection_weight,
            gate_bias=gate_bias,
            num_heads=self.num_heads,
            num_groups=self.num_groups,
            head_dim=self.head_dim,
            state_dim=self.state_dim,
        )

    def empty(
        self,
        model_dim: int,
    ) -> "Mamba2":
        in_projection = self.in_projection_config.empty(
            input_dim=model_dim,
            output_dims=(
                self.inner_dim + 2 * self.num_groups * self.state_dim,
                self.inner_dim,
                self.num_heads,
            ),
            has_biases=self.has_in_biases,
        )

        out_projection = self.out_projection_config.empty(
            self.inner_dim,
            (model_dim,),
            has_biases=self.has_out_biases,
        )

        conv = self.conv_config.empty(self.conv_dim, self.kernel_size)

        skip_connection_weight = dummy_array((self.num_heads,), in_projection.activation_precision)
        gate_bias = dummy_array((self.inner_dim,), in_projection.activation_precision)

        return Mamba2(
            self,
            in_projection=in_projection,
            conv=conv,
            out_projection=out_projection,
            skip_connection_weight=skip_connection_weight,
            gate_bias=gate_bias,
            num_heads=self.num_heads,
            num_groups=self.num_groups,
            head_dim=self.head_dim,
            state_dim=self.state_dim,
        )


class Mamba2(TokenMixerBase[Mamba2Config, SSMStateLayer]):
    in_projection: LinearBase
    conv: SeparableCausalConv
    out_projection: LinearBase

    skip_connection_weight: Float[Array, " heads"]
    gate_bias: Float[Array, " inner_channels"]

    num_heads: int = eqx.field(static=True)
    num_groups: int = eqx.field(static=True)
    head_dim: int = eqx.field(static=True)
    state_dim: int = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.in_projection.activation_precision

    @property
    def model_dim(self) -> int:
        return self.in_projection.input_dim

    @property
    def inner_dim(self) -> int:
        return self.num_heads * self.head_dim

    @property
    def positional_embedding_selector(self) -> PositionalEmbeddingSelector:
        return PositionalEmbeddingSelector.NONE

    def __post_init__(self) -> None:
        if self.skip_connection_weight.shape != (self.num_heads,):
            raise ValueError(
                f"Skip connection weight must have shape (num_heads,) = ({self.num_heads},), "
                f"got {self.skip_connection_weight.shape}",
            )
        if self.gate_bias.shape != (self.inner_dim,):
            raise ValueError(
                f"Gate bias must have shape (inner_dim,) = ({self.inner_dim},), got {self.gate_bias.shape}",
            )
        if self.num_heads % self.num_groups != 0:
            raise ValueError(
                f"Number of value heads ({self.num_heads}) must be divisible by number of groups ({self.num_groups})",
            )

    def _step(
        self,
        values: Float[Array, "heads head_dim"],
        keys: Float[Array, "groups state_dim"],
        queries: Float[Array, "groups state_dim"],
        dt_log: Float[Array, " heads"],
        state: Float[Array, "heads head_dim state_dim"],
    ) -> tuple[Float[Array, "heads head_dim"], Float[Array, "heads head_dim state_dim"]]:
        """Single-token SSM state update without scan overhead."""
        heads_per_group = self.num_heads // self.num_groups

        dt = jax.nn.softplus(dt_log)
        decay = jnp.exp(-dt)[:, None, None]
        mix = dt[:, None, None]

        keys_expanded = jnp.repeat(keys, heads_per_group, axis=0)
        queries_expanded = jnp.repeat(queries, heads_per_group, axis=0)
        values_norm = values / (dt[:, None] + 1e-8)

        input_contribution = mix * values_norm[:, :, None] * keys_expanded[:, None, :]
        new_state = decay * state + input_contribution
        output = einsum(new_state, queries_expanded, "heads head_dim state_dim, heads state_dim -> heads head_dim")

        return output, new_state

    def _decode_step(
        self,
        inputs: Float[Array, "1 channels"],
        state: SSMStateLayer,
    ) -> Mamba2Result:
        """Optimized path for single-token decode without scan machinery."""
        token = inputs[0]

        conv_in, gate, dt_log = self.in_projection(token)
        conv_out, new_conv_state = self.conv.step(conv_in, state.conv_state)
        conv_activated = self.config.activation(conv_out)

        values_flat, input_proj_flat, output_proj_flat = jnp.split(
            conv_activated,
            [self.inner_dim, self.inner_dim + self.num_groups * self.state_dim],
        )
        values = rearrange(values_flat, "(heads head_dim) -> heads head_dim", heads=self.num_heads)
        keys = rearrange(input_proj_flat, "(groups state_dim) -> groups state_dim", groups=self.num_groups)
        queries = rearrange(output_proj_flat, "(groups state_dim) -> groups state_dim", groups=self.num_groups)

        y, new_ssm_state = self._step(values, keys, queries, dt_log, state.ssm_state)

        y = y + self.skip_connection_weight[:, None] * values
        y = rearrange(y, "heads head_dim -> (heads head_dim)")
        gated = y * jax.nn.silu(gate + self.gate_bias)
        (output,) = self.out_projection(gated)

        return Mamba2Result(
            outputs=output[None, :],
            state=SSMStateLayer(new_conv_state, new_ssm_state),
        )

    def _chunked_scan(
        self,
        values: Float[Array, "suffix_tokens heads head_dim"],
        keys: Float[Array, "suffix_tokens groups state_dim"],
        queries: Float[Array, "suffix_tokens groups state_dim"],
        dt: Float[Array, "suffix_tokens heads"],
        initial_state: Float[Array, "heads head_dim state_dim"],
        chunk_size: int,
        num_steps: Int[Array, ""] | int,
        d: Float[Array, " heads"] | None = None,
        z: Float[Array, "suffix_tokens heads head_dim"] | None = None,
        z_bias: Float[Array, "heads head_dim"] | None = None,
    ) -> tuple[Float[Array, "suffix_tokens heads head_dim"], Float[Array, "heads head_dim state_dim"]]:
        """Chunked parallel scan implementing the SSD algorithm."""
        seq_len = values.shape[0]
        num_steps = jnp.asarray(num_steps, dtype=jnp.int32)

        pad_len = (chunk_size - seq_len % chunk_size) % chunk_size
        if pad_len > 0:
            values = jnp.pad(values, ((0, pad_len), (0, 0), (0, 0)))
            keys = jnp.pad(keys, ((0, pad_len), (0, 0), (0, 0)))
            queries = jnp.pad(queries, ((0, pad_len), (0, 0), (0, 0)))
            dt = jnp.pad(dt, ((0, pad_len), (0, 0)))
            if z is not None:
                z = jnp.pad(z, ((0, pad_len), (0, 0), (0, 0)))

        values_orig = values
        keys_orig = keys
        dt_orig = dt

        padded_len = values.shape[0]
        position_indices = jnp.arange(padded_len)
        valid_mask = (position_indices < num_steps).astype(values.dtype)
        values = values * valid_mask[:, None, None]
        keys = keys * valid_mask[:, None, None]

        values = rearrange(
            values,
            "(chunks chunk_size) (groups heads_per_group) head_dim"
            " -> chunks chunk_size groups heads_per_group head_dim",
            chunk_size=chunk_size,
            groups=self.num_groups,
        )
        log_decay = rearrange(
            -dt,
            "(chunks chunk_size) (groups heads_per_group) -> groups heads_per_group chunks chunk_size",
            chunk_size=chunk_size,
            groups=self.num_groups,
        )
        keys_chunked = rearrange(
            keys,
            "(chunks chunk_size) groups state_dim -> chunks chunk_size groups state_dim",
            chunk_size=chunk_size,
        )
        queries_chunked = rearrange(
            queries,
            "(chunks chunk_size) groups state_dim -> chunks chunk_size groups state_dim",
            chunk_size=chunk_size,
        )
        log_decay_cumsum = jnp.cumsum(log_decay, axis=-1)

        queries_keys_prod = einsum(
            queries_chunked,
            keys_chunked,
            "chunks query_pos groups state_dim, chunks key_pos groups state_dim -> chunks query_pos key_pos groups",
        )
        y_diag = fused_ssd_intra_chunk(log_decay_cumsum, queries_keys_prod, values)

        decay_states = jnp.exp(log_decay_cumsum[:, :, :, -1:] - log_decay_cumsum)
        states = einsum(
            keys_chunked,
            decay_states,
            values,
            "chunks chunk_size groups state_dim, groups heads_per_group chunks chunk_size,"
            " chunks chunk_size groups heads_per_group head_dim"
            " -> chunks groups heads_per_group head_dim state_dim",
        )

        initial_state_grouped = rearrange(
            initial_state,
            "(groups heads_per_group) head_dim state_dim -> groups heads_per_group head_dim state_dim",
            groups=self.num_groups,
        )
        states = jnp.concatenate([initial_state_grouped[None, ...], states], axis=0)
        log_decay_chunk_ends = jnp.pad(log_decay_cumsum[:, :, :, -1], ((0, 0), (0, 0), (1, 0)))
        decay_chunk = exp_segsum(log_decay_chunk_ends)
        new_states = einsum(
            decay_chunk,
            states,
            "groups heads_per_group out_idx chunks,"
            " chunks groups heads_per_group head_dim state_dim"
            " -> out_idx groups heads_per_group head_dim state_dim",
        )
        states = new_states[:-1]

        state_decay_out = jnp.exp(log_decay_cumsum)
        y_off = einsum(
            queries_chunked,
            states,
            state_decay_out,
            "chunks chunk_size groups state_dim,"
            " chunks groups heads_per_group head_dim state_dim,"
            " groups heads_per_group chunks chunk_size"
            " -> chunks chunk_size groups heads_per_group head_dim",
        )

        y = y_diag + y_off
        if d is not None:
            d_grouped = rearrange(d, "(groups heads_per_group) -> groups heads_per_group", groups=self.num_groups)
            y = y + d_grouped[None, None, :, :, None] * values
        y = rearrange(
            y,
            "chunks chunk_size groups heads_per_group head_dim"
            " -> (chunks chunk_size) (groups heads_per_group) head_dim",
        )

        if z is not None:
            gate = z + z_bias[None, :, :] if z_bias is not None else z
            y = y * jax.nn.silu(gate)

        y = y[:seq_len]

        new_states_flat = rearrange(
            new_states,
            "chunks groups heads_per_group head_dim state_dim -> chunks (groups heads_per_group) head_dim state_dim",
        )
        final_state = self._compute_final_state(
            values_orig,
            keys_orig,
            dt_orig,
            new_states_flat,
            num_steps,
            chunk_size,
        )

        return y, final_state

    def _compute_final_state(
        self,
        values: Float[Array, "suffix_tokens heads head_dim"],
        keys: Float[Array, "suffix_tokens groups state_dim"],
        dt: Float[Array, "suffix_tokens heads"],
        chunk_states: Float[Array, "chunks_plus_1 heads head_dim state_dim"],
        num_steps: Int[Array, ""],
        chunk_size: int,
    ) -> Float[Array, "heads head_dim state_dim"]:
        """Compute the exact final state at position num_steps using precomputed chunk_states."""
        heads_per_group = self.num_heads // self.num_groups

        chunk_idx = num_steps // chunk_size
        pos_in_chunk = num_steps % chunk_size
        chunk_start_state = jax.lax.dynamic_index_in_dim(chunk_states, chunk_idx, axis=0, keepdims=False)

        def at_boundary() -> Float[Array, "heads head_dim state_dim"]:
            return chunk_start_state

        def within_chunk() -> Float[Array, "heads head_dim state_dim"]:
            chunk_start_pos = chunk_idx * chunk_size
            values_chunk = jax.lax.dynamic_slice(
                values,
                (chunk_start_pos, 0, 0),
                (chunk_size, values.shape[1], values.shape[2]),
            )
            keys_chunk = jax.lax.dynamic_slice(
                keys,
                (chunk_start_pos, 0, 0),
                (chunk_size, keys.shape[1], keys.shape[2]),
            )
            dt_chunk = jax.lax.dynamic_slice(dt, (chunk_start_pos, 0), (chunk_size, dt.shape[1]))

            log_decay_cumsum = jnp.cumsum(-dt_chunk, axis=0)
            last_pos_idx = pos_in_chunk - 1
            log_decay_cumsum_at_last = jax.lax.dynamic_index_in_dim(
                log_decay_cumsum,
                last_pos_idx,
                axis=0,
                keepdims=False,
            )

            decayed_start = jnp.exp(log_decay_cumsum_at_last)[:, None, None] * chunk_start_state

            decay_to_last = jnp.exp(log_decay_cumsum_at_last[None, :] - log_decay_cumsum)
            mask = jnp.arange(chunk_size) <= last_pos_idx
            masked_decay = jnp.where(mask[:, None], decay_to_last, 0.0)
            keys_expanded = jnp.repeat(keys_chunk, heads_per_group, axis=1)
            input_contrib = einsum(
                masked_decay,
                keys_expanded,
                values_chunk,
                "chunk_size heads, chunk_size heads state_dim, chunk_size heads head_dim -> heads head_dim state_dim",
            )

            return decayed_start + input_contrib

        return jax.lax.cond(pos_in_chunk == 0, at_boundary, within_chunk)

    @eqx.filter_jit
    def __call__(
        self,
        inputs: Float[Array, "suffix_tokens channels"],
        positional_embeddings: PositionalEmbeddings | None,
        state: SSMStateLayer | None = None,
        return_updated_state: bool = False,
        length_without_padding: Int[Array, ""] | int | None = None,
    ) -> Mamba2Result:
        if positional_embeddings is not None:
            raise ValueError("Positional embeddings are not supported for Mamba2.")

        if state is None:
            state = SSMStateLayer.init(
                self.config.kernel_size,
                self.config.conv_dim,
                (self.num_heads, self.head_dim, self.state_dim),
                self.activation_precision,
            )

        seq_len, _ = inputs.shape

        if seq_len == 1 and return_updated_state:
            return self._decode_step(inputs, state)

        conv_inputs, gate_values, time_delta_log = vmap(self.in_projection)(inputs)

        conv_output, updated_conv_state = self.conv(
            conv_inputs,
            length_without_padding,
            state.conv_state,
            return_updated_state=return_updated_state,
        )
        conv_activated = self.config.activation(conv_output)

        x_channels, input_proj_channels, output_proj_channels = jnp.split(
            conv_activated,
            [
                self.inner_dim,
                self.inner_dim + self.num_groups * self.state_dim,
            ],
            axis=-1,
        )

        values = rearrange(
            x_channels,
            "suffix_tokens (heads head_channels) -> suffix_tokens heads head_channels",
            heads=self.num_heads,
        )
        keys = rearrange(
            input_proj_channels,
            "suffix_tokens (groups state_channels) -> suffix_tokens groups state_channels",
            groups=self.num_groups,
        )
        queries = rearrange(
            output_proj_channels,
            "suffix_tokens (groups state_channels) -> suffix_tokens groups state_channels",
            groups=self.num_groups,
        )

        if length_without_padding is None:
            length_without_padding, _ = inputs.shape

        gate_values_reshaped = rearrange(
            gate_values,
            "suffix_tokens (heads head_channels) -> suffix_tokens heads head_channels",
            heads=self.num_heads,
        )
        gate_bias_reshaped = rearrange(
            self.gate_bias,
            "(heads head_channels) -> heads head_channels",
            heads=self.num_heads,
        )

        dt = jax.nn.softplus(time_delta_log)
        ssm_outputs, final_ssm_state = self._chunked_scan(
            values,
            keys,
            queries,
            dt,
            state.ssm_state,
            self.config.chunk_size,
            length_without_padding,
            d=self.skip_connection_weight,
            z=gate_values_reshaped,
            z_bias=gate_bias_reshaped,
        )

        ssm_outputs_flat = rearrange(
            ssm_outputs,
            "suffix_tokens heads head_channels -> suffix_tokens (heads head_channels)",
        )
        (outputs,) = vmap(self.out_projection)(ssm_outputs_flat)

        if return_updated_state:
            assert updated_conv_state is not None
            updated_state = SSMStateLayer(updated_conv_state, final_ssm_state)
        else:
            updated_state = None

        return Mamba2Result(
            outputs=outputs,
            state=updated_state,
        )

    def init_static_state(self, capacity: int) -> SSMStateLayer:  # noqa: ARG002
        return SSMStateLayer.init(
            self.config.kernel_size,
            self.config.conv_dim,
            (self.num_heads, self.head_dim, self.state_dim),
            self.activation_precision,
        )

    def export_weights(self) -> ParameterTree:
        return {
            "in_projection": self.in_projection.export_weights(),
            "out_projection": self.out_projection.export_weights(),
            "conv": self.conv.export_weights(),
            "skip_connection_weight": self.skip_connection_weight,
            "gate_bias": self.gate_bias,
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        weights = require_mapping(weights)
        return replace(
            self,
            in_projection=self.in_projection.import_weights(require_tree(weights["in_projection"])),
            out_projection=self.out_projection.import_weights(require_tree(weights["out_projection"])),
            conv=self.conv.import_weights(require_tree(weights["conv"])),
            skip_connection_weight=require_array(weights["skip_connection_weight"]),
            gate_bias=require_array(weights["gate_bias"]),
        )



import functools
import itertools
import warnings
from collections.abc import Callable, Iterable
from dataclasses import dataclass

import jax
import jax.numpy as jnp
import numpy as np
from jax.errors import JaxRuntimeError
from jaxtyping import DTypeLike

from lalamo.common import LalamoWarning, get_usable_memory_from_bytes
from lalamo.models.common import InferenceConfig

type TokenSequence = list[int] | np.ndarray | jnp.ndarray

__all__ = [
    "BatchSizeEstimatingEvent",
    "decrease_batchsize_on_oom",
    "estimate_batchsize_from_bytes",
    "estimate_batchsizes_from_vram",
    "merge_small_buckets",
    "pad_keys_to_size",
    "pad_sequences",
]


@dataclass(frozen=True)
class BatchSizeEstimatingEvent:
    lo: int
    hi: int | None


def _assert_sorted(values: list[int]) -> None:
    assert all(values[i] <= values[i + 1] for i in range(len(values) - 1)), "expected sorted inputs"


def merge_small_buckets[T: TokenSequence](
    buckets: dict[int, list[tuple[int, T]]],
    batch_size_for_length: dict[int, int],
    min_batches: int = 4,
) -> dict[int, list[tuple[int, T]]]:
    # Merge buckets that are too small into larger buckets.
    # Buckets smaller than min_batches * batch_size are merged into the next larger bucket.
    # The last bucket absorbs all overflow.
    sorted_lengths = sorted(buckets.keys())
    merged: dict[int, list[tuple[int, T]]] = {}
    overflow: list[tuple[int, T]] = []

    for padded_len in sorted_lengths:
        batch_size = batch_size_for_length.get(padded_len, 1)  # note how with i's increment batch_size decreases
        items = overflow + buckets[padded_len]

        if len(items) < min_batches * batch_size:
            # the bucket is too small, push the items into a bigger one
            overflow = items
        else:
            # the bucket is big enough, keep _all_ the items and move on
            # keeping all the items avoids a funny problem with spill over into _very_ long ctx length
            merged[padded_len] = items
            overflow = []

    if overflow:
        # any leftover items go into the largest bucket
        largest_len = sorted_lengths[-1]
        merged.setdefault(largest_len, []).extend(overflow)

    return merged


def pad_sequences(
    sequences: Iterable[TokenSequence],
    shape: tuple[int, int],
    *,
    dtype: DTypeLike,
    pad_value: int = 0,
) -> jnp.ndarray:
    batch_size, seq_len = shape
    sequences_list = list(sequences)
    if len(sequences_list) > batch_size:
        raise ValueError(f"Expected at most {batch_size} sequences, got {len(sequences_list)}")

    if len(sequences_list) < batch_size:
        sequences_list.extend([jnp.array([pad_value], dtype=dtype)] * (batch_size - len(sequences_list)))

    padded = np.full((batch_size, seq_len), pad_value, dtype=dtype)
    for i, seq in enumerate(sequences_list):
        seq_arr = np.asarray(seq, dtype=dtype)
        if seq_arr.size > seq_len:
            raise ValueError(f"Sequence length {seq_arr.size} exceeds target length {seq_len}")
        padded[i, : seq_arr.size] = seq_arr

    return jnp.array(padded)


def pad_keys_to_size(keys: Iterable, size: int, *, seed: int = 0) -> jnp.ndarray:
    keys_list = list(keys)
    if len(keys_list) > size:
        raise ValueError(f"Expected at most {size} keys, got {len(keys_list)}")
    if len(keys_list) == size:
        return jnp.array(keys_list)
    dummy_keys = jax.random.split(jax.random.key(seed), size - len(keys_list))
    return jnp.concatenate([jnp.array(keys_list), dummy_keys])


def estimate_batchsizes_from_vram(
    memory_consumption_callback: Callable[[InferenceConfig], int],
    sorted_lengths: list[int],
    vram_bytes: int,
    inference_config: InferenceConfig,
) -> dict[int, int]:
    _assert_sorted(sorted_lengths)
    assert len(sorted_lengths) > 0
    usable_memory = get_usable_memory_from_bytes(vram_bytes)

    def memory_consumption(bs: int, seq_len: int) -> int:
        config = InferenceConfig(
            max_output_length=inference_config.max_output_length,
            padded_length=seq_len,
            num_top_logits_to_return=inference_config.num_top_logits_to_return,
            batch_size=bs,
        )
        return memory_consumption_callback(config)

    result: dict[int, int] = {}

    first_seq_len = sorted_lengths[0]
    bs = estimate_batchsize_from_bytes(functools.partial(memory_consumption, seq_len=sorted_lengths[0]), usable_memory)
    result[first_seq_len] = bs
    for seq_len in sorted_lengths[1:]:
        while bs > 1 and memory_consumption(bs, seq_len) > usable_memory:
            bs = max(1, int(bs * 0.8))

        result[seq_len] = bs

    return result


def estimate_batchsize_from_bytes(
    memory_per_batchsize_callback: Callable[[int], int],
    target_mem_bytes: int,
    progress: Callable[[BatchSizeEstimatingEvent], None] | None = None,
) -> int:
    lo = 0
    hi = 0
    for candidate_exp in itertools.count():
        lo = hi
        hi = 4**candidate_exp

        if progress is not None:
            progress(BatchSizeEstimatingEvent(lo, None))
        if target_mem_bytes < memory_per_batchsize_callback(hi):
            break

    while hi - lo > 1:
        mid = (lo + hi) // 2

        if progress is not None:
            progress(BatchSizeEstimatingEvent(lo, hi))
        if target_mem_bytes < memory_per_batchsize_callback(mid):
            hi = mid
        else:
            lo = mid

    return lo


def decrease_batchsize_on_oom[T](
    fn: Callable[[int], Iterable[T]],
    starting_batch_size: int,
) -> Iterable[T]:
    first_batch_completed = False
    effective_batch_size = starting_batch_size

    while True:
        try:
            for result in fn(effective_batch_size):
                yield result

                # as soon as we yielded we are not allowed to retry anymore
                # to make sure we don't ever miss/duplicate outputs
                first_batch_completed = True
            break
        except JaxRuntimeError as e:
            if first_batch_completed:
                raise
            # because OOM's sometimes generate stuff that won't be garbage collected,
            # we need to be very aggressive with decreasing batchsize here
            new_bs = max(int(0.7 * effective_batch_size - 1), 1)
            if new_bs == 1 and effective_batch_size == 1:
                raise
            warnings.warn(
                f"OOM detected ({e}). Reducing batch size {effective_batch_size} -> {new_bs}.",
                LalamoWarning,
                stacklevel=3,
            )
            effective_batch_size = new_bs



from lalamo.model_import.model_configs import (
    HFBonsaiConfig,
    HFQwen2Config,
    HFQwen3Config,
    HFQwen3NextConfig,
    HFQwen35Config,
)
from lalamo.quantization import QuantizationMode

from .common import ConfigMap, FileSpec, ModelSpec, UseCase, WeightsType

__all__ = ["QWEN_MODELS"]


QWEN25 = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-0.5B-Instruct",
        size="0.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-0.5B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-1.5B-Instruct",
        size="1.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-1.5B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-3B-Instruct",
        size="3B",
        quantization=None,
        repo="Qwen/Qwen2.5-3B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-7B-Instruct",
        size="7B",
        quantization=None,
        repo="Qwen/Qwen2.5-7B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-14B-Instruct",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen2.5-14B-Instruct",
        config_type=HFQwen2Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5",
        name="Qwen2.5-32B-Instruct",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen2.5-32B-Instruct",
        config_type=HFQwen2Config,
    ),
]


QWEN25_CODER = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-0.5B-Instruct",
        size="0.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-0.5B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-1.5B-Instruct",
        size="1.5B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-1.5B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-3B-Instruct",
        size="3B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-3B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-7B-Instruct",
        size="7B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-7B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-14B-Instruct",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-14B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-32B-Instruct",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-32B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen2.5-Coder",
        name="Qwen2.5-Coder-32B-Instruct",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen2.5-Coder-32B-Instruct",
        config_type=HFQwen2Config,
        use_cases=(UseCase.CODE,),
    ),
]


QWEN3 = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-0.6B",
        size="0.6B",
        quantization=None,
        repo="Qwen/Qwen3-0.6B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-0.6B-MLX-4bit",
        size="0.6B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-0.6B-MLX-4bit",
        config_type=HFQwen3Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3-0.6B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3-0.6B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3-0.6B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-1.7B",
        size="1.7B",
        quantization=None,
        repo="Qwen/Qwen3-1.7B",
        config_type=HFQwen3Config,
        weights_type=WeightsType.SAFETENSORS,
        use_cases=tuple(),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-4B",
        size="4B",
        quantization=None,
        repo="Qwen/Qwen3-4B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-4B-AWQ",
        size="4B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-4B-AWQ",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-4B-MLX-4bit",
        size="4B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-4B-MLX-4bit",
        config_type=HFQwen3Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3-4B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3-4B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3-4B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-8B",
        size="8B",
        quantization=None,
        repo="Qwen/Qwen3-8B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-8B-AWQ",
        size="8B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-8B-AWQ",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-8B-MLX-4bit",
        size="8B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-8B-MLX-4bit",
        config_type=HFQwen3Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3-8B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3-8B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3-8B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-14B",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen3-14B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-14B-AWQ",
        size="14B",
        quantization=None,
        repo="Qwen/Qwen3-14B-AWQ",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-32B",
        size="32B",
        quantization=None,
        repo="Qwen/Qwen3-32B",
        config_type=HFQwen3Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3",
        name="Qwen3-32B-AWQ",
        size="32B",
        quantization=QuantizationMode.UINT4,
        repo="Qwen/Qwen3-32B-AWQ",
        config_type=HFQwen3Config,
    ),
]

QWEN3_NEXT = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3-Next",
        name="Qwen3-Next-80B-A3B-Instruct",
        size="80B",
        quantization=None,
        repo="Qwen/Qwen3-Next-80B-A3B-Instruct",
        config_type=HFQwen3NextConfig,
    ),
    # TODO(knyazer): uncomment and test when sharding is in
    # ModelSpec(
    #     vendor="Alibaba",
    #     family="Qwen3-Next",
    #     name="Qwen3-Next-80B-A3B-Instruct-4bit",
    #     size="80B",
    #     quantization=QuantizationMode.UINT4,
    #     repo="mlx-community/Qwen3-Next-80B-A3B-Instruct-4bit",
    #     config_type=HFQwen3NextConfig,
    #     configs=ConfigMap(
    #         tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3-Next-80B-A3B-Instruct"),
    #         tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3-Next-80B-A3B-Instruct"),
    #         generation_config=FileSpec("generation_config.json", "Qwen/Qwen3-Next-80B-A3B-Instruct"),
    #     ),
    # ),
]

QWEN35 = [
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-0.8B",
        size="0.8B",
        quantization=None,
        repo="Qwen/Qwen3.5-0.8B",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-0.8B-MLX-4bit",
        size="0.8B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/Qwen3.5-0.8B-MLX-4bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-0.8B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-0.8B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-0.8B-MLX-8bit",
        size="0.8B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/Qwen3.5-0.8B-MLX-8bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-0.8B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-0.8B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-2B",
        size="2B",
        quantization=None,
        repo="Qwen/Qwen3.5-2B",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-2B-MLX-4bit",
        size="2B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/Qwen3.5-2B-MLX-4bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-2B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-2B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-2B-MLX-8bit",
        size="2B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/Qwen3.5-2B-MLX-8bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-2B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-2B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-4B",
        size="4B",
        quantization=None,
        repo="Qwen/Qwen3.5-4B",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-4B-MLX-4bit",
        size="4B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/Qwen3.5-4B-MLX-4bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-4B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-4B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-4B-MLX-8bit",
        size="4B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/Qwen3.5-4B-MLX-8bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-4B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-4B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-9B",
        size="9B",
        quantization=None,
        repo="Qwen/Qwen3.5-9B",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-9B-MLX-4bit",
        size="9B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/Qwen3.5-9B-MLX-4bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-9B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-9B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-9B-MLX-8bit",
        size="9B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/Qwen3.5-9B-MLX-8bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-9B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-9B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-27B",
        size="27B",
        quantization=None,
        repo="Qwen/Qwen3.5-27B",
        config_type=HFQwen35Config,
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-27B-MLX-4bit",
        size="27B",
        quantization=QuantizationMode.UINT4,
        repo="mlx-community/Qwen3.5-27B-4bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-27B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-27B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
    ModelSpec(
        vendor="Alibaba",
        family="Qwen3.5",
        name="Qwen3.5-27B-MLX-8bit",
        size="27B",
        quantization=QuantizationMode.UINT8,
        repo="mlx-community/Qwen3.5-27B-8bit",
        config_type=HFQwen35Config,
        configs=ConfigMap(
            tokenizer=FileSpec("tokenizer.json", "Qwen/Qwen3.5-27B"),
            tokenizer_config=FileSpec("tokenizer_config.json", "Qwen/Qwen3.5-27B"),
            generation_config=FileSpec("generation_config.json", "Qwen/Qwen3.5-27B"),
        ),
    ),
]

QWEN3_PARD = [
    ModelSpec(
        vendor="AMD",
        family="PARD",
        name="PARD-Qwen3-0.6B",
        size="0.6B",
        quantization=None,
        repo="amd/PARD-Qwen3-0.6B",
        config_type=HFQwen3Config,
    ),
]

BONSAI = [
    ModelSpec(
        vendor="PrismML",
        family="Bonsai",
        name="Bonsai-4B-mlx-1bit",
        size="4B",
        quantization=QuantizationMode.UINT1,
        repo="prism-ml/Bonsai-4B-mlx-1bit",
        config_type=HFBonsaiConfig,
        configs=ConfigMap(
            chat_template=FileSpec("chat_template.jinja"),
        ),
    ),
]

QWEN_MODELS = QWEN25 + QWEN25_CODER + QWEN3 + QWEN3_NEXT + QWEN35 + QWEN3_PARD + BONSAI



"""Lalamo implementation of NanoCodec modules from NVIDIA NeMo.

Reference: https://github.com/NVIDIA-NeMo/NeMo/blob/v2.3.0/nemo/collections/tts/modules/audio_codec_modules.py
Reference: Mentzer et al., Finite Scalar Quantization: VQ-VAE Made Simple (https://arxiv.org/abs/2309.15505v1)
"""

from collections.abc import Mapping, Sequence
from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree, require_array, require_tree
from lalamo.modules.audio.common_modules import (
    CausalConv1d,
    CausalConv1dConfig,
    CausalTransposeConv1d,
    CausalTransposeConv1dConfig,
    Snake1d,
    Snake1dConfig,
)
from lalamo.modules.common import LalamoModule

__all__ = [
    "CausalHiFiGANDecoder",
    "CausalHiFiGANDecoderConfig",
    "CausalTransposeConv1d",
    "CausalTransposeConv1dConfig",
    "FiniteScalarQuantizer",
    "FiniteScalarQuantizerConfig",
    "GroupFiniteScalarQuantizer",
    "GroupFiniteScalarQuantizerConfig",
    "HalfSnake",
    "HalfSnakeConfig",
    "HiFiGANResBlock",
    "HiFiGANResBlockConfig",
    "HiFiGANResLayer",
    "HiFiGANResLayerConfig",
    "ResidualBlock",
    "ResidualBlockConfig",
]


@dataclass(frozen=True)
class FiniteScalarQuantizerConfig:
    num_levels: tuple[int, ...]
    eps: float
    precision: DTypeLike

    @property
    def dim(self) -> int:
        """Dimension of the input/output vectors."""
        return len(self.num_levels)

    @property
    def codebook_size(self) -> int:
        """Size of the implicit codebook."""
        result = 1
        for level in self.num_levels:
            result *= level
        return result

    def empty(self) -> "FiniteScalarQuantizer":
        """Create quantizer with buffer arrays.

        Note: FSQ has no learnable weights, only precomputed buffers.
        """
        num_levels = jnp.array(self.num_levels, dtype=jnp.int32)
        # dim_base_index = cumprod([1, levels[0], levels[0]*levels[1], ...])
        # Used to convert per-dimension indices to a single codebook index
        dim_base_index = jnp.cumprod(jnp.concatenate([jnp.array([1], dtype=jnp.int32), num_levels[:-1]]))
        return FiniteScalarQuantizer(
            config=self,
            num_levels_buffer=num_levels,
            dim_base_index=dim_base_index,
        )

    def random_init(
        self,
        *,
        key: PRNGKeyArray,  # noqa: ARG002 (unused - FSQ has no learnable weights)
    ) -> "FiniteScalarQuantizer":
        return self.empty()


class FiniteScalarQuantizer(LalamoModule[FiniteScalarQuantizerConfig]):
    """Finite Scalar Quantizer.

    Quantizes each element of the input vector independently into a number of levels.
    Uses tanh compression and rounding with straight-through estimator.

    Reference: Mentzer et al., Finite Scalar Quantization: VQ-VAE Made Simple
    """

    num_levels_buffer: Int[Array, " dim"]
    dim_base_index: Int[Array, " dim"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def dim(self) -> int:
        return self.config.dim

    @property
    def codebook_size(self) -> int:
        return self.config.codebook_size

    def _compress(self, inputs: Float[Array, "batch dim seq"]) -> Float[Array, "batch dim seq"]:
        """Apply tanh compression to map continuous values to quantization range.

        The compression ensures values are bounded within the quantization levels.
        For a dimension with L levels, outputs are mapped to range [-(L-1)/2, (L-1)/2]
        with appropriate handling for even/odd number of levels.
        """
        num_levels = self.num_levels_buffer.astype(self.config.precision)

        # Output scale: (num_levels - 1) / 2
        # Scaled down slightly to avoid rounding issues at boundaries
        output_scale = (num_levels - 1) / 2
        output_scale = output_scale * (1 - self.config.eps)

        # Offset for even number of levels (shifts grid by 0.5)
        output_offset = jnp.where(self.num_levels_buffer % 2 == 0, 0.5, 0.0)

        # Input shift for even levels (pre-tanh adjustment)
        input_shift = jnp.tan(output_offset / output_scale)

        # Apply compression: scale * tanh(input + shift) - offset
        output = (
            output_scale[None, :, None] * jnp.tanh(inputs + input_shift[None, :, None]) - output_offset[None, :, None]
        )
        return output

    def _round_ste(self, inputs: Float[Array, "batch dim seq"]) -> Float[Array, "batch dim seq"]:
        """Round to nearest integer with straight-through estimator."""
        rounded = jnp.round(inputs)
        return inputs + jax.lax.stop_gradient(rounded - inputs)

    def _inputs_to_codes(self, inputs: Float[Array, "batch dim seq"]) -> Float[Array, "batch dim seq"]:
        """Convert continuous inputs to quantized codes normalized to [-1, 1].

        Steps:
        1. Compress inputs using tanh to bound them
        2. Round to nearest integer level
        3. Normalize to [-1, 1] range
        """
        compressed = self._compress(inputs)
        codes = self._round_ste(compressed)
        # Normalize to [-1, 1]
        scale = (self.num_levels_buffer // 2).astype(self.config.precision)
        codes = codes / scale[None, :, None]
        return codes

    def _codes_to_nonnegative(self, codes: Float[Array, "batch dim seq"]) -> Float[Array, "batch dim seq"]:
        """Convert codes centered around zero to nonnegative integer indices."""
        scale = (self.num_levels_buffer // 2).astype(self.config.precision)
        offset = scale
        return scale[None, :, None] * codes + offset[None, :, None]

    def _nonnegative_to_codes(self, codes_nonneg: Float[Array, "dim"]) -> Float[Array, "dim"]:
        """Convert nonnegative indices back to codes centered around zero."""
        scale = (self.num_levels_buffer // 2).astype(self.config.precision)
        offset = scale
        return (codes_nonneg - offset) / scale

    def _codes_to_indices(self, codes: Float[Array, "batch dim seq"]) -> Int[Array, "batch seq"]:
        """Convert per-dimension code vectors to single indices.

        Uses the dim_base_index to compute: sum(code_d * base_d) for d in dimensions
        """
        nonneg = self._codes_to_nonnegative(codes)
        # Sum over dimensions weighted by base index
        indices = jnp.sum(nonneg * self.dim_base_index[None, :, None].astype(self.config.precision), axis=1)
        return indices.astype(jnp.int32)

    def _indices_to_codes(self, index: Int[Array, " 1"]) -> Float[Array, " dim"]:
        """Convert single indices to per-dimension code vectors.
        Reverses the indexing: code_d = (index // base_d) % levels_d
        """

        codes_nonnegative = (index // self.dim_base_index) % self.num_levels_buffer
        return self._nonnegative_to_codes(codes_nonnegative.astype(self.config.precision))

    def encode(
        self,
        inputs: Float[Array, "batch dim seq"],
    ) -> Int[Array, "batch seq"]:
        """Encode continuous inputs to discrete indices."""
        codes = self._inputs_to_codes(inputs)
        indices = self._codes_to_indices(codes)
        # Add codebook dimension for compatibility with RVQ API
        return indices

    def decode(
        self,
        indices: Int[Array, " seq"],
    ) -> Float[Array, "seq dim"]:
        """Decode discrete indices back to continuous code vectors."""
        return jax.vmap(self._indices_to_codes)(indices)

    def __call__(
        self,
        inputs: Float[Array, "batch seq"],
    ) -> Float[Array, "batch seq dim"]:
        """
        Forward pass: dequantize batch of input indices vectors to continuous representation.
        """
        return jax.vmap(self.decode, in_axes=0)(inputs)

    def export_weights(self) -> ParameterTree[Array]:
        return {
            "num_levels_buffer": self.num_levels_buffer,
            "dim_base_index": self.dim_base_index,
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            num_levels_buffer=require_array(weights["num_levels_buffer"]),
            dim_base_index=require_array(weights["dim_base_index"]),
        )


@dataclass(frozen=True)
class GroupFiniteScalarQuantizerConfig:
    num_groups: int
    quantizer_config: FiniteScalarQuantizerConfig

    @property
    def precision(self) -> DTypeLike:
        return self.quantizer_config.precision

    @property
    def codebook_dim_per_group(self) -> int:
        """Dimension per group (number of levels dimensions)."""
        return self.quantizer_config.dim

    @property
    def codebook_dim(self) -> int:
        """Total input dimension (groups * dim_per_group)."""
        return self.codebook_dim_per_group * self.num_groups

    @property
    def codebook_size_per_group(self) -> int:
        """Codebook size for each group (product of levels)."""
        return self.quantizer_config.codebook_size

    def empty(self) -> "GroupFiniteScalarQuantizer":
        quantizers = tuple(self.quantizer_config.empty() for _ in range(self.num_groups))
        return GroupFiniteScalarQuantizer(config=self, quantizers=quantizers)

    def random_init(
        self,
        *,
        key: PRNGKeyArray,  # noqa: ARG002 (unused - FSQ has no learnable weights)
    ) -> "GroupFiniteScalarQuantizer":
        return self.empty()


class GroupFiniteScalarQuantizer(LalamoModule[GroupFiniteScalarQuantizerConfig]):
    """Group Finite Scalar Quantizer.

    Splits the input vector into groups and applies FSQ on each group separately.
    This enables hierarchical/group-wise quantization where each group produces
    independent indices.
    """

    quantizers: tuple[FiniteScalarQuantizer, ...]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def num_groups(self) -> int:
        return self.config.num_groups

    @property
    def codebook_dim(self) -> int:
        return self.config.codebook_dim

    @property
    def codebook_dim_per_group(self) -> int:
        return self.config.codebook_dim_per_group

    def encode(
        self,
        inputs: Float[Array, "batch channels seq"],
    ) -> Int[Array, "num_groups batch seq"]:
        """Encode inputs to indices for each group."""
        # Split input along channel dimension into groups
        inputs_grouped = jnp.split(inputs, self.num_groups, axis=1)

        indices_list = []
        for in_group, quantizer in zip(inputs_grouped, self.quantizers, strict=True):
            idx = quantizer.encode(in_group)[None, :, :]  # [1, batch, seq]
            indices_list.append(idx)

        # Concatenate along codebook/group dimension
        return jnp.concatenate(indices_list, axis=0)  # [num_groups, batch, seq]

    def decode(
        self,
        indices: Int[Array, "batch seq num_groups"],
    ) -> Float[Array, "batch seq channels"]:
        """Decode batch of indices vectors back to continuous representation."""
        # # Split indices along group dimension
        indices_grouped = jnp.split(indices, self.num_groups, axis=2)

        dequantized_list = []
        for idx_group, quantizer in zip(indices_grouped, self.quantizers, strict=True):
            deq = quantizer(idx_group)
            dequantized_list.append(deq)

        return jnp.concatenate(dequantized_list, axis=2)

    def __call__(
        self,
        inputs: Float[Array, "batch seq num_groups"],
    ) -> Float[Array, "batch seq channels"]:
        return self.decode(inputs)

    def export_weights(self) -> ParameterTree[Array]:
        return {
            "quantizers": [q.export_weights() for q in self.quantizers],
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        quantizer_weights = weights["quantizers"]
        assert isinstance(quantizer_weights, Sequence)

        new_quantizers = tuple(
            q.import_weights(require_tree(w)) for q, w in zip(self.quantizers, quantizer_weights, strict=True)
        )
        return replace(self, quantizers=new_quantizers)


@dataclass(frozen=True)
class HalfSnakeConfig:
    snake_config: Snake1dConfig
    leaky_relu_negative_slope: float

    @property
    def precision(self) -> DTypeLike:
        return self.snake_config.precision

    def empty(self, channels: int) -> "HalfSnake":
        snake_channels = channels // 2
        snake = self.snake_config.empty(snake_channels)
        return HalfSnake(config=self, snake=snake, total_channels=channels)

    def random_init(
        self,
        channels: int,
        *,
        key: PRNGKeyArray,  # noqa: ARG002 (unused - Snake uses ones initialization)
    ) -> "HalfSnake":
        snake_channels = channels // 2
        snake = self.snake_config.random_init(snake_channels)
        return HalfSnake(config=self, snake=snake, total_channels=channels)


class HalfSnake(LalamoModule[HalfSnakeConfig]):
    """HalfSnake activation function.

    Applies Snake activation to the first half of channels and
    LeakyReLU to the second half.
    """

    snake: Snake1d
    total_channels: int = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def channels(self) -> int:
        return self.total_channels

    @property
    def snake_channels(self) -> int:
        return self.total_channels // 2

    def __call__(
        self,
        x: Float[Array, "batch sequence channels"],
    ) -> Float[Array, "batch sequence channels"]:
        # Split along channel dimension (axis=2 for NSC format)
        snake_input = x[:, :, : self.snake_channels]
        lrelu_input = x[:, :, self.snake_channels :]

        # Apply Snake to first half
        snake_out = self.snake(snake_input)

        # Apply LeakyReLU to second half
        negative_slope = self.config.leaky_relu_negative_slope
        lrelu_out = jnp.where(lrelu_input >= 0, lrelu_input, negative_slope * lrelu_input)

        # Concatenate along channel dimension
        return jnp.concatenate([snake_out, lrelu_out], axis=2)

    def export_weights(self) -> ParameterTree[Array]:
        return {"snake": self.snake.export_weights()}

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        snake_weights = require_tree(weights["snake"])
        new_snake = self.snake.import_weights(snake_weights)
        return replace(self, snake=new_snake)


@dataclass(frozen=True)
class ResidualBlockConfig:
    """Configuration for ResidualBlock.

    A residual block that applies activation → conv → activation → conv + residual.
    Used in HiFi-GAN decoder architecture.
    """

    activation_config: HalfSnakeConfig
    conv_config: CausalConv1dConfig

    @property
    def precision(self) -> DTypeLike:
        return self.conv_config.precision

    def empty(
        self,
        channels: int,
        kernel_size: int,
        dilation: int,
    ) -> "ResidualBlock":
        input_activation = self.activation_config.empty(channels)
        skip_activation = self.activation_config.empty(channels)

        input_conv = self.conv_config.empty(
            in_channels=channels,
            out_channels=channels,
            kernel_size=kernel_size,
            dilation=dilation,
        )
        skip_conv = self.conv_config.empty(
            in_channels=channels,
            out_channels=channels,
            kernel_size=kernel_size,
            dilation=1,
        )

        return ResidualBlock(
            config=self,
            input_activation=input_activation,
            skip_activation=skip_activation,
            input_conv=input_conv,
            skip_conv=skip_conv,
        )

    def random_init(
        self,
        channels: int,
        kernel_size: int,
        dilation: int,
        *,
        key: PRNGKeyArray,
    ) -> "ResidualBlock":
        input_act_key, skip_act_key, input_conv_key, skip_conv_key = jax.random.split(key, 4)

        input_activation = self.activation_config.random_init(channels, key=input_act_key)
        skip_activation = self.activation_config.random_init(channels, key=skip_act_key)

        input_conv = self.conv_config.random_init(
            in_channels=channels,
            out_channels=channels,
            kernel_size=kernel_size,
            dilation=dilation,
            key=input_conv_key,
        )
        skip_conv = self.conv_config.random_init(
            in_channels=channels,
            out_channels=channels,
            kernel_size=kernel_size,
            dilation=1,
            key=skip_conv_key,
        )

        return ResidualBlock(
            config=self,
            input_activation=input_activation,
            skip_activation=skip_activation,
            input_conv=input_conv,
            skip_conv=skip_conv,
        )


class ResidualBlock(LalamoModule[ResidualBlockConfig]):
    """Residual block for HiFi-GAN decoder.
    Applies activation → conv → activation → conv + residual
    """

    input_activation: HalfSnake
    skip_activation: HalfSnake
    input_conv: CausalConv1d
    skip_conv: CausalConv1d

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def channels(self) -> int:
        return self.input_conv.in_channels

    def __call__(
        self,
        x: Float[Array, "batch sequence channels"],
    ) -> Float[Array, "batch sequence channels"]:
        # activation → conv → activation → conv
        conv_input = self.input_activation(x)
        skip_input = self.input_conv(conv_input)
        skip_input = self.skip_activation(skip_input)
        res = self.skip_conv(skip_input)

        # Residual connection
        return x + res

    def export_weights(self) -> ParameterTree[Array]:
        return {
            "input_activation": self.input_activation.export_weights(),
            "skip_activation": self.skip_activation.export_weights(),
            "input_conv": self.input_conv.export_weights(),
            "skip_conv": self.skip_conv.export_weights(),
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)

        input_act_weights = weights["input_activation"]
        skip_act_weights = weights["skip_activation"]
        input_conv_weights = weights["input_conv"]
        skip_conv_weights = weights["skip_conv"]

        return replace(
            self,
            input_activation=self.input_activation.import_weights(require_tree(input_act_weights)),
            skip_activation=self.skip_activation.import_weights(require_tree(skip_act_weights)),
            input_conv=self.input_conv.import_weights(require_tree(input_conv_weights)),
            skip_conv=self.skip_conv.import_weights(require_tree(skip_conv_weights)),
        )


@dataclass(frozen=True)
class HiFiGANResBlockConfig:
    """Configuration for HiFiGANResBlock.

    Wraps multiple ResidualBlocks with different dilations, applied sequentially.
    Used in HiFi-GAN decoder architecture.
    """

    residual_block_config: ResidualBlockConfig

    @property
    def precision(self) -> DTypeLike:
        return self.residual_block_config.precision

    def empty(
        self,
        channels: int,
        kernel_size: int,
        dilations: tuple[int, ...],
    ) -> "HiFiGANResBlock":
        res_blocks = tuple(
            self.residual_block_config.empty(
                channels=channels,
                kernel_size=kernel_size,
                dilation=dilation,
            )
            for dilation in dilations
        )
        return HiFiGANResBlock(config=self, res_blocks=res_blocks)

    def random_init(
        self,
        channels: int,
        kernel_size: int,
        dilations: tuple[int, ...],
        *,
        key: PRNGKeyArray,
    ) -> "HiFiGANResBlock":
        keys = jax.random.split(key, len(dilations))
        res_blocks = tuple(
            self.residual_block_config.random_init(
                channels=channels,
                kernel_size=kernel_size,
                dilation=dilation,
                key=block_key,
            )
            for dilation, block_key in zip(dilations, keys, strict=True)
        )
        return HiFiGANResBlock(config=self, res_blocks=res_blocks)


class HiFiGANResBlock(LalamoModule[HiFiGANResBlockConfig]):
    """HiFiGAN residual block wrapper.
    Creates multiple ResidualBlocks with different dilations and applies them sequentially.
    """

    res_blocks: tuple[ResidualBlock, ...]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def channels(self) -> int:
        return self.res_blocks[0].channels

    def __call__(
        self,
        x: Float[Array, "batch sequence channels"],
    ) -> Float[Array, "batch sequence channels"]:
        out = x
        for res_block in self.res_blocks:
            out = res_block(out)
        return out

    def export_weights(self) -> ParameterTree[Array]:
        return {
            "res_blocks": [block.export_weights() for block in self.res_blocks],
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        block_weights = weights["res_blocks"]
        assert isinstance(block_weights, Sequence)

        new_blocks = tuple(
            block.import_weights(require_tree(w)) for block, w in zip(self.res_blocks, block_weights, strict=True)
        )
        return replace(self, res_blocks=new_blocks)


@dataclass(frozen=True)
class HiFiGANResLayerConfig:
    """Configuration for HiFiGANResLayer.

    Creates multiple HiFiGANResBlocks with different kernel sizes.
    The outputs of all blocks are averaged.
    """

    hifigan_res_block_config: HiFiGANResBlockConfig

    @property
    def precision(self) -> DTypeLike:
        return self.hifigan_res_block_config.precision

    def empty(
        self,
        channels: int,
        kernel_sizes: tuple[int, ...],
        dilations: tuple[int, ...],
    ) -> "HiFiGANResLayer":
        res_blocks = tuple(
            self.hifigan_res_block_config.empty(
                channels=channels,
                kernel_size=kernel_size,
                dilations=dilations,
            )
            for kernel_size in kernel_sizes
        )
        return HiFiGANResLayer(config=self, res_blocks=res_blocks)

    def random_init(
        self,
        channels: int,
        kernel_sizes: tuple[int, ...],
        dilations: tuple[int, ...],
        *,
        key: PRNGKeyArray,
    ) -> "HiFiGANResLayer":
        keys = jax.random.split(key, len(kernel_sizes))
        res_blocks = tuple(
            self.hifigan_res_block_config.random_init(
                channels=channels,
                kernel_size=kernel_size,
                dilations=dilations,
                key=block_key,
            )
            for kernel_size, block_key in zip(kernel_sizes, keys, strict=True)
        )
        return HiFiGANResLayer(config=self, res_blocks=res_blocks)


class HiFiGANResLayer(LalamoModule[HiFiGANResLayerConfig]):
    """HiFiGAN residual layer.

    Creates multiple HiFiGANResBlocks with different kernel sizes.
    Each block processes the same input, and their outputs are averaged.
    """

    res_blocks: tuple[HiFiGANResBlock, ...]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def channels(self) -> int:
        return self.res_blocks[0].channels

    def __call__(
        self,
        x: Float[Array, "batch sequence channels"],
    ) -> Float[Array, "batch sequence channels"]:
        residuals = jnp.stack([res_block(x) for res_block in self.res_blocks], axis=0)
        return jnp.mean(residuals, axis=0)

    def export_weights(self) -> ParameterTree[Array]:
        return {
            "res_blocks": [block.export_weights() for block in self.res_blocks],
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        block_weights = weights["res_blocks"]
        assert isinstance(block_weights, Sequence)

        new_blocks = tuple(
            block.import_weights(require_tree(w)) for block, w in zip(self.res_blocks, block_weights, strict=True)
        )
        return replace(self, res_blocks=new_blocks)


@dataclass(frozen=True)
class CausalHiFiGANDecoderConfig:
    activation_config: HalfSnakeConfig
    pre_conv_config: CausalConv1dConfig
    transpose_conv_config: CausalTransposeConv1dConfig
    res_layer_config: HiFiGANResLayerConfig
    post_conv_config: CausalConv1dConfig

    @property
    def precision(self) -> DTypeLike:
        return self.pre_conv_config.precision

    def empty(
        self,
        input_dim: int,
        base_channels: int,
        up_sample_rates: tuple[int, ...],
        in_kernel_size: int,
        out_kernel_size: int,
        resblock_kernel_sizes: tuple[int, ...],
        resblock_dilations: tuple[int, ...],
    ) -> "CausalHiFiGANDecoder":
        # Pre-conv: input_dim -> base_channels
        pre_conv = self.pre_conv_config.empty(
            in_channels=input_dim,
            out_channels=base_channels,
            kernel_size=in_kernel_size,
        )

        # Build stages: each stage has activation, upsample conv, res_layer
        activations = []
        upsample_convs = []
        res_layers = []

        in_channels = base_channels
        for up_sample_rate in up_sample_rates:
            out_channels = in_channels // 2
            kernel_size = 2 * up_sample_rate

            # Activation before upsample
            activation = self.activation_config.empty(in_channels)
            activations.append(activation)

            # Upsample conv (groups=out_channels for depthwise-like operation)
            upsample_conv = self.transpose_conv_config.empty(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=up_sample_rate,
                groups=out_channels,
            )
            upsample_convs.append(upsample_conv)

            # Res layer (at reduced channel count)
            res_layer = self.res_layer_config.empty(
                channels=out_channels,
                kernel_sizes=resblock_kernel_sizes,
                dilations=resblock_dilations,
            )
            res_layers.append(res_layer)

            in_channels = out_channels

        # Post activation and conv
        post_activation = self.activation_config.empty(in_channels)
        post_conv = self.post_conv_config.empty(
            in_channels=in_channels,
            out_channels=1,
            kernel_size=out_kernel_size,
        )

        return CausalHiFiGANDecoder(
            config=self,
            pre_conv=pre_conv,
            activations=tuple(activations),
            upsample_convs=tuple(upsample_convs),
            res_layers=tuple(res_layers),
            post_activation=post_activation,
            post_conv=post_conv,
            up_sample_rates=up_sample_rates,
        )

    def random_init(
        self,
        input_dim: int,
        base_channels: int,
        up_sample_rates: tuple[int, ...],
        in_kernel_size: int,
        out_kernel_size: int,
        resblock_kernel_sizes: tuple[int, ...],
        resblock_dilations: tuple[int, ...],
        *,
        key: PRNGKeyArray,
    ) -> "CausalHiFiGANDecoder":
        num_stages = len(up_sample_rates)
        # pre_conv, post_activation, post_conv + 3 per stage (activation, upsample, res_layer)
        num_keys = 3 + 3 * num_stages
        keys = jax.random.split(key, num_keys)
        key_idx = 0

        pre_conv = self.pre_conv_config.random_init(
            in_channels=input_dim,
            out_channels=base_channels,
            kernel_size=in_kernel_size,
            key=keys[key_idx],
        )
        key_idx += 1

        activations = []
        upsample_convs = []
        res_layers = []

        in_channels = base_channels
        for up_sample_rate in up_sample_rates:
            out_channels = in_channels // 2
            kernel_size = 2 * up_sample_rate

            activation = self.activation_config.random_init(in_channels, key=keys[key_idx])
            key_idx += 1
            activations.append(activation)

            upsample_conv = self.transpose_conv_config.random_init(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=up_sample_rate,
                groups=out_channels,
                key=keys[key_idx],
            )
            key_idx += 1
            upsample_convs.append(upsample_conv)

            res_layer = self.res_layer_config.random_init(
                channels=out_channels,
                kernel_sizes=resblock_kernel_sizes,
                dilations=resblock_dilations,
                key=keys[key_idx],
            )
            key_idx += 1
            res_layers.append(res_layer)

            in_channels = out_channels

        post_activation = self.activation_config.random_init(in_channels, key=keys[key_idx])
        key_idx += 1
        post_conv = self.post_conv_config.random_init(
            in_channels=in_channels,
            out_channels=1,
            kernel_size=out_kernel_size,
            key=keys[key_idx],
        )

        return CausalHiFiGANDecoder(
            config=self,
            pre_conv=pre_conv,
            activations=tuple(activations),
            upsample_convs=tuple(upsample_convs),
            res_layers=tuple(res_layers),
            post_activation=post_activation,
            post_conv=post_conv,
            up_sample_rates=up_sample_rates,
        )


class CausalHiFiGANDecoder(LalamoModule[CausalHiFiGANDecoderConfig]):
    """Causal HiFi-GAN decoder for audio generation.

    Converts quantized representations to audio waveforms using:
    1. Pre-conv to expand to base_channels
    2. Multiple stages of: activation -> upsample -> res_layer
    3. Post-conv to produce single-channel audio
    4. Tanh over result as substitute for clipping

    Returns audio waveform in (batch, audio_length) format
    """

    pre_conv: CausalConv1d
    activations: tuple[HalfSnake, ...]
    upsample_convs: tuple[CausalTransposeConv1d, ...]
    res_layers: tuple[HiFiGANResLayer, ...]
    post_activation: HalfSnake
    post_conv: CausalConv1d
    up_sample_rates: tuple[int, ...] = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    def __call__(
        self,
        x: Float[Array, "batch sequence channels"],
    ) -> Float[Array, "batch audio_length"]:
        # Pre-conv: [B, S, input_dim] -> [B, S, base_channels]
        out = self.pre_conv(x)

        # Upsample stages
        for activation, upsample_conv, res_layer in zip(
            self.activations,
            self.upsample_convs,
            self.res_layers,
            strict=True,
        ):
            out = activation(out)
            out = upsample_conv(out)
            out = res_layer(out)

        out = self.post_activation(out)
        out = self.post_conv(out)  # [B, T_audio, 1]

        audio = jnp.tanh(out)
        audio = audio[:, :, 0]  # [B, T_audio]

        return audio

    def export_weights(self) -> ParameterTree[Array]:
        return {
            "pre_conv": self.pre_conv.export_weights(),
            "activations": [act.export_weights() for act in self.activations],
            "upsample_convs": [conv.export_weights() for conv in self.upsample_convs],
            "res_layers": [layer.export_weights() for layer in self.res_layers],
            "post_activation": self.post_activation.export_weights(),
            "post_conv": self.post_conv.export_weights(),
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)

        pre_conv_weights = require_tree(weights["pre_conv"])
        activations_weights = weights["activations"]
        upsample_convs_weights = weights["upsample_convs"]
        res_layers_weights = weights["res_layers"]
        post_activation_weights = require_tree(weights["post_activation"])
        post_conv_weights = require_tree(weights["post_conv"])

        assert isinstance(activations_weights, Sequence)
        assert isinstance(upsample_convs_weights, Sequence)
        assert isinstance(res_layers_weights, Sequence)

        new_activations = tuple(
            act.import_weights(require_tree(w)) for act, w in zip(self.activations, activations_weights, strict=True)
        )
        new_upsample_convs = tuple(
            conv.import_weights(require_tree(w))
            for conv, w in zip(self.upsample_convs, upsample_convs_weights, strict=True)
        )
        new_res_layers = tuple(
            layer.import_weights(require_tree(w)) for layer, w in zip(self.res_layers, res_layers_weights, strict=True)
        )

        return replace(
            self,
            pre_conv=self.pre_conv.import_weights(pre_conv_weights),
            activations=new_activations,
            upsample_convs=new_upsample_convs,
            res_layers=new_res_layers,
            post_activation=self.post_activation.import_weights(post_activation_weights),
            post_conv=self.post_conv.import_weights(post_conv_weights),
        )



"""Foreign config for NanoCodec TTS models from NVIDIA NeMo."""

import json
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Self

import numpy as np
from jax import numpy as jnp
from jaxtyping import Array, DTypeLike

from lalamo.model_import.loaders.common import load_parameters
from lalamo.model_import.loaders.nanocodec_loaders import load_nanocodec
from lalamo.model_import.model_configs import ForeignTTSConfig
from lalamo.modules import LalamoModule, TTSConfig, TTSModel, VocoderConfig
from lalamo.modules.audio.common_modules import CausalConv1dConfig
from lalamo.modules.audio.fishaudio.fishaudio_modules import Snake1dConfig
from lalamo.modules.audio.nanocodec.audio_decoding import NanoCodec, NanoCodecConfig
from lalamo.modules.audio.nanocodec.nanocodec_consts import (
    DEFAULT_AUDIO_DECODER_INPUT_CONV_SIZE,
    DEFAULT_AUDIO_DECODER_OUTPUT_CONV_SIZE,
    DEFAULT_AUDIO_DECODER_RESBLOCK_DILATIONS,
    DEFAULT_AUDIO_DECODER_RESBLOCK_KERNEL_SIZES,
    DEFAULT_FSQ_EPS,
    DEFAULT_LEAKY_RELU_NEGATIVE_SLOPE,
)
from lalamo.modules.audio.nanocodec.nanocodec_modules import (
    CausalHiFiGANDecoderConfig,
    CausalTransposeConv1dConfig,
    FiniteScalarQuantizerConfig,
    GroupFiniteScalarQuantizerConfig,
    HalfSnakeConfig,
    HiFiGANResBlockConfig,
    HiFiGANResLayerConfig,
    ResidualBlockConfig,
)
from lalamo.modules.audio.nanocodec.stub_text_decoder import StubTextDecoder, StubTextDecoderConfig

__all__ = ["NanoCodecForeignConfig"]


def _require_key(config: Mapping, key: str, context: str) -> Any:  # noqa: ANN401
    """Get a required key from config or raise a clear error."""
    if key not in config:
        raise ValueError(f"Required key '{key}' not found in {context} config")
    return config[key]


@dataclass(frozen=True)
class NanoCodecForeignConfig(ForeignTTSConfig):
    """Foreign config for NanoCodec models from NVIDIA NeMo.

    Fields are flattened from the NeMo YAML config sections:
    - sample_rate (top-level)
    - vector_quantizer.num_groups, vector_quantizer.num_levels_per_group
    - audio_decoder.base_channels, audio_decoder.up_sample_rates, etc.
    """

    # Top-level
    samplerate: int

    # vector_quantizer section
    num_groups: int
    num_levels_per_group: tuple[int, ...]

    # audio_decoder section
    base_channels: int
    up_sample_rates: tuple[int, ...]
    in_kernel_size: int
    out_kernel_size: int
    resblock_kernel_sizes: tuple[int, ...]
    resblock_dilation_sizes: tuple[int, ...]

    def to_tts_config(
        self,
        context_length: int | None,  # noqa: ARG002
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,  # noqa: ARG002
    ) -> TTSConfig:
        fsq_config = FiniteScalarQuantizerConfig(
            num_levels=self.num_levels_per_group,
            eps=DEFAULT_FSQ_EPS,
            precision=activation_precision,
        )

        quantizer_config = GroupFiniteScalarQuantizerConfig(
            num_groups=self.num_groups,
            quantizer_config=fsq_config,
        )

        snake_config = Snake1dConfig(precision=activation_precision)
        activation_config = HalfSnakeConfig(
            snake_config=snake_config,
            leaky_relu_negative_slope=DEFAULT_LEAKY_RELU_NEGATIVE_SLOPE,
        )
        conv_config = CausalConv1dConfig(precision=activation_precision, has_biases=True)
        transpose_conv_config = CausalTransposeConv1dConfig(precision=activation_precision, has_biases=True)

        residual_block_config = ResidualBlockConfig(
            activation_config=activation_config,
            conv_config=conv_config,
        )
        hifigan_res_block_config = HiFiGANResBlockConfig(residual_block_config=residual_block_config)
        res_layer_config = HiFiGANResLayerConfig(hifigan_res_block_config=hifigan_res_block_config)

        decoder_config = CausalHiFiGANDecoderConfig(
            activation_config=activation_config,
            pre_conv_config=conv_config,
            transpose_conv_config=transpose_conv_config,
            res_layer_config=res_layer_config,
            post_conv_config=conv_config,
        )

        num_codebooks = self.num_groups
        codebook_size = int(np.prod(self.num_levels_per_group))

        audio_decoder_config = NanoCodecConfig(
            precision=activation_precision,
            quantizer_config=quantizer_config,
            decoder_config=decoder_config,
            samplerate=self.samplerate,
            base_channels=self.base_channels,
            up_sample_rates=self.up_sample_rates,
            in_kernel_size=self.in_kernel_size,
            out_kernel_size=self.out_kernel_size,
            resblock_kernel_sizes=self.resblock_kernel_sizes,
            resblock_dilations=self.resblock_dilation_sizes,
        )

        text_decoder_config = StubTextDecoderConfig(
            num_codebooks=num_codebooks,
            codebook_size=codebook_size,
            precision=activation_precision,
        )

        return TTSConfig(
            text_decoder_config=text_decoder_config,
            audio_decoder_config=audio_decoder_config,
            vocoder_config=VocoderConfig(),
            activation_precision=activation_precision,
        )

    def _load_weights(
        self,
        model: LalamoModule,
        weights_dict: Mapping[str, Array],
    ) -> LalamoModule:
        assert isinstance(model, TTSModel)
        assert isinstance(model.text_decoder, StubTextDecoder)
        assert isinstance(model.audio_decoder, NanoCodec)

        loaded_audio_decoder = load_nanocodec(model.audio_decoder, weights_dict)

        return load_parameters(
            lambda m: (m.audio_decoder,),
            model,
            (loaded_audio_decoder,),
        )

    @classmethod
    def from_json(cls, json_path: Path | str) -> Self:
        json_path = Path(json_path)
        with open(json_path) as f:
            config = json.load(f)
        return cls.from_nemo_config(config)

    @classmethod
    def from_nemo_config(cls, nemo_config: Mapping) -> Self:
        """Create from a NeMo YAML config dict (extracted from .nemo archive)."""
        quantizer_config = _require_key(nemo_config, "vector_quantizer", "nemo")
        audio_decoder_config = _require_key(nemo_config, "audio_decoder", "nemo")
        assert isinstance(quantizer_config, Mapping)
        assert isinstance(audio_decoder_config, Mapping)

        return cls(
            samplerate=int(_require_key(nemo_config, "sample_rate", "nemo")),
            num_groups=int(_require_key(quantizer_config, "num_groups", "vector_quantizer")),
            num_levels_per_group=tuple(_require_key(quantizer_config, "num_levels_per_group", "vector_quantizer")),
            base_channels=int(_require_key(audio_decoder_config, "base_channels", "audio_decoder")),
            up_sample_rates=tuple(_require_key(audio_decoder_config, "up_sample_rates", "audio_decoder")),
            in_kernel_size=int(audio_decoder_config.get("in_kernel_size", DEFAULT_AUDIO_DECODER_INPUT_CONV_SIZE)),
            out_kernel_size=int(audio_decoder_config.get("out_kernel_size", DEFAULT_AUDIO_DECODER_OUTPUT_CONV_SIZE)),
            resblock_kernel_sizes=audio_decoder_config.get(
                "resblock_kernel_sizes",
                DEFAULT_AUDIO_DECODER_RESBLOCK_KERNEL_SIZES,
            ),
            resblock_dilation_sizes=audio_decoder_config.get(
                "resblock_dilation_sizes",
                DEFAULT_AUDIO_DECODER_RESBLOCK_DILATIONS,
            ),
        )

    @property
    def default_precision(self) -> DTypeLike:
        return jnp.float32



from dataclasses import dataclass
from typing import Self

from jax import numpy as jnp
from jaxtyping import Array, DTypeLike, Float

from lalamo.common import ParameterTree
from lalamo.modules.common import LalamoModule


@dataclass(frozen=True)
class NoopVocoderConfig:
    def empty(self) -> "NoopVocoder":
        return NoopVocoder(config=self)


class NoopVocoder(LalamoModule[NoopVocoderConfig]):
    # Used in models where vocoder is absent and there is no additional
    # conversion from audio-features space to audio waveform space

    @property
    def activation_precision(self) -> DTypeLike:
        return jnp.float32

    def export_weights(self) -> ParameterTree[Array]:
        return {}

    def import_weights(
        self,
        weights: ParameterTree[Array],  # noqa: ARG002
    ) -> Self:
        return self

    def __call__(
        self,
        audio_features: Float[Array, "batch audio_channels samples"],
    ) -> Float[Array, "batch audio_channels samples"]:
        return audio_features


VocoderConfig = NoopVocoderConfig
Vocoder = NoopVocoder



from lalamo.model_import.model_configs.huggingface.fishaudio import FishAudioConfig

from .common import ConfigMap, FileSpec, ModelSpec, ModelType, WeightsType

__all__ = ["FISHAUDIO_TTS_MODELS"]

DIRECT_CHAT_TEMPLATE = """
{% for message in messages %}<|{{message.style}}|><|{{message.speaker_id}}|>{{message.content}}{% endfor %}
"""

FISHAUDIO_TTS_MODELS = [
    ModelSpec(
        vendor="FishAudio",
        family="openaudio",
        name="s1-mini",
        size="0.8B",
        quantization=None,
        repo="fishaudio/s1-mini",
        config_type=FishAudioConfig,
        weights_type=WeightsType.TORCH,
        configs=ConfigMap(
            chat_template=DIRECT_CHAT_TEMPLATE,
            generation_config=None,
            tokenizer=FileSpec("tokenizer.tiktoken"),
        ),
        model_type=ModelType.TTS_MODEL,
    ),
]
